import abc
from abc import ABC
from functools import cache
from typing import Generic, Optional, Union, Tuple, Collection, Any, Dict, Set, TypeVar

import numpy as np
from pytensor import tensor as at

from sakkara.model.base import ModelComponent
from sakkara.model.math_op import MathOpBase
from sakkara.relation.groupset import GroupSet
from sakkara.relation.nodepair import NodePair

S = TypeVar('S', bound=Any)
T = TypeVar('T', bound=ModelComponent)


class Composable(MathOpBase, ABC, Generic[S, T]):
    """
    Base class for a component that can be built with underlying subcomponents



    :param name: Name of the corresponding variable to register in PyMC.

    :param group: Group of which the component is defined for.

    :param members: Subset of members of column the component is defined for.

    :param subcomponents: Dict of underlying ModelComponent objects.

    """

    def __init__(self, name: Optional[str], group: Optional[Union[str, Tuple[str, ...]]],
                 members: Optional[Collection[Any]], subcomponents: Dict[S, T]):
        super().__init__()
        self.name = name
        self.members = members
        self.member_nodes = None
        self.group = (group,) if isinstance(group, str) else group
        self.subcomponents = subcomponents
        self.group_node = None
        self.components_node = None

    @abc.abstractmethod
    def __getitem__(self, item: Any) -> T:
        raise NotImplementedError

    def get_name(self) -> Optional[str]:
        return self.name

    def set_name(self, name: str) -> None:
        self.name = name

    def build_components(self, groupset: GroupSet) -> None:
        for param_name, component in self.subcomponents.items():
            if component.get_name() is None:
                component.set_name(f'{param_name}_{self.get_name()}')
            if component.variable is None:
                component.build(groupset)

    @abc.abstractmethod
    def get_built_components(self) -> Dict[S, at.Variable]:
        raise NotImplementedError

    def build_group_nodes(self, groupset: GroupSet) -> None:
        if 0 < len(self.subcomponents):
            self.components_node = next(iter(self.subcomponents.values())).node

            for component in self.subcomponents.values():
                self.components_node = NodePair(self.components_node, component.node).reduced_repr()
        else:
            self.components_node = groupset['global']

        if self.group is not None:
            self.group_node = groupset[self.group[0]]
            for column in self.group[1:]:
                self.group_node = NodePair(self.group_node, groupset[column]).reduced_repr()

            self.node = NodePair(self.group_node, self.components_node).reduced_repr()
        else:
            self.group_node = self.components_node
            self.node = self.components_node

    def build_member_nodes(self):
        if self.members is None:
            self.member_nodes = self.group_node.get_members().ravel()
        else:
            self.members = {m if isinstance(m, Tuple) else (m,) for m in self.members}
            key_match = np.vectorize(lambda node: node.get_key() in self.members)
            self.member_nodes = self.group_node.get_members().ravel()[
                key_match(self.group_node.get_members().ravel())]

    def build_node(self, groupset: GroupSet):
        self.build_group_nodes(groupset)
        self.build_member_nodes()

    def clear(self):
        self.variable = None
        self.node = None
        for c in self.subcomponents.values():
            c.clear()

    @cache
    def retrieve_groups(self) -> Set[str]:
        group = set()
        for k, v in self.subcomponents.items():
            parent_groups = v.retrieve_groups()
            group = group.union(parent_groups)
        if self.group is not None:
            group = group.union(self.group)

        return group

    def dims(self):
        return tuple(map(str, self.node.representation()))



from copy import deepcopy
from typing import Any, Union, Tuple, Optional, Set

import numpy as np

from sakkara.model.math_op import MathOpBase
from sakkara.relation.groupset import GroupSet
from sakkara.relation.nodepair import NodePair


class FixedValueComponent(MathOpBase):
    """
    Class for non-random fixed values that appear in model that are assigned to a group.

    :param value: Value to wrap into the component.
    :param name: Name of the corresponding variable to register in PyMC.
    :param group: Group of which the component is defined for.

    """
    def __init__(self, value: Any, name: Optional[str] = None, group: Union[str, Tuple[str, ...]] = 'global'):
        super().__init__()
        self.group = (group,) if isinstance(group, str) else group
        self.name = name
        self.values = np.array(value).reshape(1) if isinstance(value, (float, int)) else value

    def set_name(self, name: str) -> None:
        self.name = name

    def get_name(self) -> Optional[str]:
        return self.name if self.name is not None else 'fixed'

    def clear(self):
        self.variable = None
        self.node = None

    def prebuild(self, groupset: GroupSet) -> None:
        pass

    def build_node(self, groupset: GroupSet) -> None:
        self.node = groupset[self.group[0]]
        for column in self.group[1:]:
            self.node = NodePair(groupset[column], self.node).reduced_repr()

    def build_variable(self) -> None:
        self.variable = deepcopy(self.values)

    def retrieve_groups(self) -> Set[str]:
        return set(self.group)



from abc import ABC
from typing import Union, Tuple, Collection, Any

from sakkara.model.composable.hierarchical.wrapper import ComponentWrapper


class MemberWrapper(ComponentWrapper[ComponentWrapper], ABC):
    """
    Helper class for wrapping component that is defined only for a subset of members in a group.


    :param component: ModelComponent object to wrap.
    :param group: Group which the component should be wrapped into.
    :param members: Members of the group that the component is defined for.
    :param name: Name of the component to appear in PyMC.
    """

    def __init__(self, component: ComponentWrapper, group: Union[str, Tuple[str, ...]] = None,
                 members=Collection[Any], name: str = None):
        super().__init__(component, group, members, name)

    def __getitem__(self, item) -> ComponentWrapper:
        return self.subcomponents[item]



import operator
from abc import ABC
from itertools import chain
from typing import Callable, Any, Set, Optional

import pytensor.tensor as pt

from sakkara.model.base import ModelComponent
from sakkara.relation.nodepair import NodePair
from sakkara.relation.groupset import GroupSet


class FunctionComponent(ModelComponent, ABC):
    """
    Class for intermediate states of mathematical operations between components

    :param fct: Function to evaluate

    :param \**kwargs: See below

    :Arguments:
        * *arg* (``ModelComponent``) --
          Arguments passed to fct. If object does not inherit ``ModelComponent``, you may wrap it with :class:`sakkara.model.FixedValueComponent`

    :Keyword Arguments:
        * *kwarg* (``ModelComponent``) --
          Keyword argument passed to fct. If object does not inherit ``ModelComponent``, you may wrap it with :class:`sakkara.model.FixedValueComponent`
    """

    def __init__(self, fct: Callable[[Any, ...], Any], *args: ModelComponent, **kwargs: ModelComponent):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.fct = fct

    def get_name(self) -> Optional[str]:
        for comp in chain(self.args, self.kwargs.values()):
            comp_name = comp.get_name()
            if comp_name is None:
                return None

        return str(self.fct)

    def set_name(self, name: str) -> None:
        for i, comp in enumerate(self.args):
            if comp.get_name() is None:
                comp.set_name(name + '_' + str(self.fct) + '_arg' + str(i))

        for k, comp in self.kwargs.items():
            if comp.get_name() is None:
                comp.set_name(name + '_' + str(self.fct) + '_' + k)

    def clear(self):
        self.variable = None
        for comp in chain(self.args, self.kwargs.values()):
            comp.clear()

    def prebuild(self, groupset: GroupSet) -> None:
        unbuilt = [c for c in chain(self.args, self.kwargs.values())]

        all_args = len(self.args) + len(self.kwargs)
        counter = 0
        while 0 < len(unbuilt):
            component = unbuilt.pop(0)
            if component.get_name() is None:
                # Other components may be needed to be built first
                if counter < all_args * (1 + all_args) / 2:
                    unbuilt.append(component)
                else:
                    # Did not help to build other component first, name must be defined for this explicitly
                    raise ValueError('All arguments to must be named')
            elif component.variable is None:
                component.build(groupset)
            counter += 1

    def build_node(self, groupset: GroupSet) -> None:
        self.node = None
        for comp in chain(self.args, self.kwargs.values()):
            if self.node is None:
                self.node = comp.node
            else:
                self.node = NodePair(self.node, comp.node).reduced_repr()

    def build_variable(self) -> None:
        def get_mapped_variable(component: ModelComponent) -> pt.TensorVariable:
            if component.node.get_members().shape == (1,):
                return component.variable
            mapping = self.node.map_to(component.node)
            return component.variable[mapping]

        mapped_args = tuple(map(get_mapped_variable, self.args))
        mapped_kwargs = dict({k: get_mapped_variable(v) for k, v in self.kwargs.items()})
        self.variable = self.fct(*mapped_args, **mapped_kwargs)

    def retrieve_groups(self) -> Set[str]:
        group = set()
        for component in chain(self.args, self.kwargs.values()):
            group = group.union(component.retrieve_groups())
        return group

    @staticmethod
    def math_op(fct: Callable, left: Any, right: Any) -> ModelComponent:
        if isinstance(left, ModelComponent) and isinstance(right, ModelComponent):
            return FunctionComponent(fct, left, right)
        if isinstance(left, ModelComponent):
            return FunctionComponent(lambda x: fct(x, right), left)
        return FunctionComponent(lambda x: fct(left, x), right)

    def __add__(self, other: Any) -> ModelComponent:
        return self.math_op(operator.add, self, other)

    def __sub__(self, other: Any) -> ModelComponent:
        return self.math_op(operator.sub, self, other)

    def __radd__(self, other):
        return self.math_op(operator.add, other, self)

    def __rsub__(self, other):
        return self.math_op(operator.sub, other, self)

    def __mul__(self, other: Any) -> ModelComponent:
        return self.math_op(operator.mul, self, other)

    def __rmul__(self, other: Any) -> ModelComponent:
        return self.math_op(operator.add, other, self)

    def __truediv__(self, other: Any) -> ModelComponent:
        return self.math_op(operator.truediv, self, other)

    def __rtruediv__(self, other: Any):
        return self.math_op(operator.truediv, other, self)

    def __neg__(self):
        return FunctionComponent(operator.neg, self)



from dataclasses import dataclass
from typing import Dict, List, Set, Any

import numpy as np
import pandas as pd

from sakkara.relation.cellnode import CellNode
from sakkara.relation.atomicnode import AtomicNode


@dataclass(frozen=True)
class GroupSet:
    """
    Set of group nodes used for PyMC model creation
    """
    groups: Dict[str, AtomicNode]

    def __getitem__(self, item: str):
        return self.groups[item]

    def coords(self) -> Dict[str, np.ndarray]:
        coords_dict = {}
        for k, v in self.groups.items():
            names = np.empty(len(v.get_members()), dtype=object)
            for i, m in enumerate(v.get_members()):
                names[i] = m.get_key()[0]
            coords_dict[k] = names
        return coords_dict


def get_parent_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Create a matrix of parent mappings between columns in a dataframe
    Parameters
    ----------
    df: Dataframe (with C columns) to base parent mappings of.

    Returns
    -------
    Dataframe of size CxC with parent mappings. Rows are sorted from highest to lowest hierarchy level. Element (i,j)
    True means that j is a parent to i.
    """
    groups = list(df.columns)
    n_uniques = df.nunique(axis=0)
    counts_df = pd.DataFrame(index=groups, columns=groups, data=False)
    for i in range(len(groups)):
        counts_df.loc[groups[i], groups[:i] + groups[i + 1:]] = np.logical_and(
            df.groupby(groups[i]).nunique().max() == 1,
            n_uniques[groups[i]] > n_uniques[groups[:i] + groups[i + 1:]])

    counts_df['rank'] = counts_df.sum(axis=1)
    counts_df.sort_values(by='rank', inplace=True, ascending=True)
    return counts_df.loc[:, df.columns]


def get_member_parents(child_key: str, df: pd.DataFrame, parents: Set[AtomicNode]) -> Dict[Any, Set[CellNode]]:
    """
    Add member parents of the (child) group
    Parameters
    ----------
    child_key: Name of the (child) group
    df: Original dataframe
    parents: All composite groups that is a parent to the composite group with group_name
    """
    member_parents = {k: set() for k in list(df[child_key])}
    if len(parents) > 0:
        mappings = df.groupby(child_key).first().reset_index()

        for parent in parents:
            parent_mappings = mappings[parent.get_key()[0]]
            for parent_member in parent.get_members():
                for matched_child_key in mappings.loc[parent_mappings == parent_member.get_key()[0], child_key]:
                    if matched_child_key not in member_parents:
                        member_parents[matched_child_key] = set()
                    member_parents[matched_child_key].add(parent_member)

    return member_parents


def get_twin_df(df):
    groups = list(df.columns)
    counts_df = pd.DataFrame(index=groups, columns=groups, data=False)
    for i in range(len(groups)):
        counts_df.loc[groups[i], groups[:i] + groups[i + 1:]] = df.groupby(groups[i]).nunique().max() == 1
    return np.logical_and(counts_df, counts_df.T)


def create_column_node(column: str, parent_names: List[str], df: pd.DataFrame,
                       groups: Dict[Any, AtomicNode]) -> AtomicNode:
    """
    Init a single composite group.

    Parameters
    ----------
    column: Name of the composite group
    parent_names: Names of the parents
    df: Orginal dataframe
    groups: Dict of names mapped to the pre-created parents

    Returns
    -------

    """
    parents = set(groups[pn] for pn in parent_names)
    selection_df = df.loc[:, [column] + parent_names]

    member_parents = get_member_parents(column, selection_df, parents)

    members = np.empty(len(member_parents), dtype=object)
    for i, member_name in enumerate(df[column].unique()):
        new_member = CellNode(member_name, member_parents[member_name])
        # Add new member as child to member of parent
        for member_parent in member_parents[member_name]:
            member_parent.add_child(new_member)
        members[i] = new_member

    # Create group node
    new_node = AtomicNode(column, members, parents)
    for parent in parents:
        parent.add_child(new_node)

    return new_node


def init(df: pd.DataFrame) -> GroupSet:
    """
    Init a group set from a dataframe
    :param df: DataFrame containing only the group columns
    :return: GroupSet created from the input DataFrame
    """
    tmp_df = df.copy()

    groups = dict()

    parent_df = get_parent_df(df)
    for i, (group_name, is_parent) in enumerate(parent_df.iterrows()):
        groups[str(group_name)] = create_column_node(str(group_name), list(is_parent.index[is_parent]), tmp_df, groups)

    twin_df = get_twin_df(df)
    for group_name, is_twin in twin_df.iterrows():
        for name, twin_i in is_twin[is_twin].items():
            groups[str(group_name)].add_twin(groups[str(name)])
            for member, twin_member in zip(groups[group_name].get_members(), groups[str(name)].get_members()):
                member.add_twin(twin_member)

    return GroupSet(groups)



from abc import ABC
from typing import Callable, Optional, Union, Tuple, Any, Iterable

from sakkara.model.base import ModelComponent
from sakkara.model.composable.base import T
from sakkara.model.fixed.base import FixedValueComponent
from sakkara.model.composable.hierarchical.base import HierarchicalComponent


class DistributionComponent(HierarchicalComponent[T], ABC):
    """
    Class for components whose variable is generated by a PyMC distribution


    :param generator: PyMC callable for the distribution to use
    :type generator: Callable

    :param name: Name of the corresponding variable to register in PyMC, defaults to None
    :type name: str

    :param group: Group of which the component is defined for, defaults to None
    :type group: Union[str, Tuple[str, ...]]

    :param members: Subset of members of column the component is defined for, defaults to None
    :type members: Iterable[Any]
    
    :param \**subcomponents: Underlying components/objects passed as parameters to PyMC distribution, should correspond to keyword of `generator`

    **Example**

    .. highlight:: python
    .. code-block:: python

        import pymc as pm
        from sakkara.model import DistributionComponent as DC
        sigma_comp = DC(pm.HalfNormal)
        n = DC(pm.Normal, sigma=sigma_comp)

    """
    def __init__(self, generator: Callable, name: Optional[str] = None, group: Union[str, Tuple[str, ...]] = None,
                 members: Optional[Iterable[Any]] = None, **subcomponents: Any):
        super().__init__(name, group, members,
                         subcomponents={k: v if isinstance(v, ModelComponent) else FixedValueComponent(v) for k, v in
                                        subcomponents.items()})
        self.generator = generator

    def build_variable(self) -> None:
        self.variable = self.generator(self.name, **self.get_built_components(), shape=self.node.get_members().shape,
                                       dims=self.dims())



from abc import ABC
from typing import Dict, Union

import pandas as pd
import numpy.typing as npt

from sakkara.model.fixed.base import FixedValueComponent


class SeriesComponent(FixedValueComponent, ABC):
    """
    Helper component for wrapping Pandas Series objects

    :param data: Series to wrap as a component.
    :param name: Name of the corresponding variable to register in PyMC.
    :param group: Group of which the component is defined for.
    """
    def __init__(self, data: Union[npt.NDArray, pd.Series], name: str, group: str = 'obs'):
        super().__init__(data.values if isinstance(data, pd.Series) else data, name, group)


def data_components(df: pd.DataFrame) -> Dict[str, SeriesComponent]:
    """
    Generate :class:`SeriesComponent` objects from a :class:`pandas.DataFrame`

    :param df: DataFrame to generate components from.


    :return: Dictionary of {<column name in DataFrame>: :class:`SeriesComponent`}

    """
    return {k: SeriesComponent(df[k], k) for k in df}



from abc import ABC
from typing import Union, Tuple, Optional, Collection, Any

from sakkara.model.composable.base import T
from sakkara.model.composable.hierarchical.base import HierarchicalComponent


class ComponentWrapper(HierarchicalComponent[T], ABC):
    """
    Base class for wrapping a single component.


    :param component: ModelComponent object to wrap.
    :param group: Group which the component should be wrapped into.
    :param members: Members of the group that the component is defined for.
    :param name: Name of the component to appear in PyMC.
    """
    def __init__(self, component: T, group: Union[str, Tuple[str, ...]],
                 members: Optional[Collection[Any]], name: str = None):
        super().__init__(name, group, members, {'component': component})

    def build_variable(self) -> None:
        self.variable = self.get_built_components()['component']



from abc import ABC
from typing import Tuple, Union, Optional, Collection, Any, Dict

import pymc as pm
from pytensor import tensor as pt
from numpy import typing as npt

from sakkara.model.base import ModelComponent
from sakkara.model.composable.base import Composable
from sakkara.model.fixed.base import FixedValueComponent
from sakkara.model.composable.group.groupwrapper import GroupWrapper
from sakkara.model.composable.group.memberwrapper import MemberWrapper
from sakkara.relation.groupset import GroupSet


class GroupComponent(Composable[Tuple[str, ...], MemberWrapper], ABC):
    """
    Class for specifying components for each member of a group individually

    :param group: Group of which the component is defined for.
    :param membercomponents: Dictionary with key indicating member (corresponding to DataFrame value) and value its
            corresponding value (ModelComponent or other)
    :param name: Name of the corresponding variable to register in PyMC.
    """
    def __init__(self, group: Union[str, Tuple[str, ...]], members: Optional[Collection[Any]] = None,
                 name: str = None, membercomponents: Dict[Any, Any] = None):
        super().__init__(name, group, members, {})
        if membercomponents is not None:
            for k, v in membercomponents.items():
                self.add(k, v)

    def __getitem__(self, item: Any):
        return self.subcomponents[item if isinstance(item, tuple) else (item,)]['component']['component']

    def add(self, member: Any, component: Union[float, npt.NDArray, ModelComponent]) -> None:
        """
        Add component for a member to the GroupComponent

        :param member: Key for the group member, corresponding to Dataframe value.
        :param component: Value or component of the given member.
        """
        if not isinstance(component, ModelComponent):
            component = FixedValueComponent(component)
        key = member if isinstance(member, tuple) else (member,)
        self.subcomponents[key] = MemberWrapper(GroupWrapper(component, name=member),
                                                group=self.group,
                                                members=[member],
                                                name=member)

    def set_group(self):
        component_groups = set()
        for member_comp in self.subcomponents.values():
            component_groups = component_groups.union(member_comp['component']['component'].retrieve_groups())

        for member_comp in self.subcomponents.values():
            member_comp['component'].group = tuple(component_groups)

    def prebuild(self, groupset: GroupSet) -> None:
        self.set_group()
        self.build_components(groupset)

    def get_built_components(self) -> Dict[Tuple[str, ...], pt.Variable]:
        return {k: v.variable for k, v in self.subcomponents.items()}

    def build_variable(self) -> None:
        undefined_members = list(
            m for m in self.group_node.get_members() if m.get_key() not in list(self.get_built_components().keys()))
        if 0 < len(undefined_members):
            raise KeyError('All members of group must be defined to build concat variable')

        pt_tensor = pt.stack([self.get_built_components()[m.get_key()] for m in self.group_node.get_members().ravel()])
        self.variable = pm.Deterministic(name=self.name,
                                         var=pt_tensor.reshape(self.node.get_members().shape),
                                         dims=self.dims())



import numpy as np
import pandas as pd
import pymc as pm

from sakkara.model.base import ModelComponent
from sakkara.relation.groupset import init


def build(df: pd.DataFrame, component: ModelComponent):
    """
    Build a complete PyMC model based on a single :class:`ModelComponent` (typically :class:`Likelihood`). Sakkara
    will trace all underlying components, and their respective groupings, necessary for creating the model.

    :param df: :class:`pandas.DataFrame` containing columns defining groups used among :class:`ModelComponent` objects.

    :param component: :class:`ModelComponent` object (of the lowest hierarchy, typically a :class:`Likelihood`) to init creation of PyMC model
        from.

    :return: A PyMC model generated by the dataframe and component.

    :rtype: :class:`pymc.Model`

    """
    tmp_df = df.copy()
    tmp_df.loc[:, 'global'] = 'global'
    tmp_df.loc[:, 'obs'] = np.arange(len(df))

    groups = component.retrieve_groups().union({'global', 'obs'})
    groupset = init(tmp_df.loc[:, list(groups)])

    with pm.Model(coords=groupset.coords()) as model:
        component.build(groupset)
    return model



from abc import ABC
from typing import Union, Tuple

from sakkara.model.base import ModelComponent
from sakkara.model.composable.hierarchical.wrapper import ComponentWrapper


class GroupWrapper(ComponentWrapper[ModelComponent], ABC):
    """
    Helper class for wrapping component in a (potentially) different group than originally defined.

    :param component: ModelComponent object to wrap.
    :param group: Group which the component should be wrapped into.
    :param name: Name of the component to appear in PyMC.
    """

    def __init__(self, component: ModelComponent, group: Union[str, Tuple[str, ...]] = None, name: str = None):
        super().__init__(component, group, None, name)

    def __getitem__(self, item) -> ModelComponent:
        return self.subcomponents[item]



import abc
from typing import Set, Optional, Any

from sakkara.relation.groupset import GroupSet


class ModelComponent:
    """
        Abstract class for all model components
    """

    def __init__(self):
        self.node = None
        self.variable = None

    @abc.abstractmethod
    def get_name(self) -> Optional[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def set_name(self, name: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self) -> None:
        """
        Clear variable and group for this and underlying components
        """
        raise NotImplementedError

    @abc.abstractmethod
    def build_variable(self) -> None:
        """
        Build the variable, performed after prebuild and build_group.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_groups(self) -> Set[str]:
        """
        Retrieve group names for this and underlying components
        """
        raise NotImplementedError

    @abc.abstractmethod
    def prebuild(self, groupset: GroupSet) -> None:
        """
        All operations to be performed before building group and variable, e.g., building the underlying components.

        :param groupset: Groups to be used for building all components of the model.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def build_node(self, groupset: GroupSet) -> None:
        """
        Build the group of this component, performed after prebuild.

        :param groupset: Groups to be used for building all components of the model.

        """
        raise NotImplementedError

    def build(self, groupset: GroupSet) -> None:
        self.prebuild(groupset)
        self.build_node(groupset)
        self.build_variable()

    @abc.abstractmethod
    def __add__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __sub__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __mul__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __rmul__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __truediv__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __rtruediv__(self, other: Any) -> 'ModelComponent':
        raise NotImplementedError

    @abc.abstractmethod
    def __neg__(self) -> 'ModelComponent':
        raise NotImplementedError



import warnings
from abc import ABC
from typing import Callable, Any, Dict, Union, Tuple

import numpy as np

from sakkara.model.fixed.base import FixedValueComponent
from sakkara.model.fixed.data import SeriesComponent
from sakkara.model.composable.group.component import GroupComponent
from sakkara.model.composable.hierarchical.distribution import DistributionComponent


class Likelihood(DistributionComponent, ABC):
    """
    Component for a likelihood distribution, e.g, with observed data.

    :param generator: PyMC callable for distribution to use.
    :param observed: Data to input as observed keyword in PyMC.
    :param name: Name of the corresponding variable to register in PyMC.
    :param group: Group of which the component is defined for.
    :param nan_param_mask: Masked distribution parameters to use for rows with `Nan`, must be defined for each keyword argument entered. Required if there are `Nan` in observed.
    :param nan_data_mask: Masked observed value to use for rows with `Nan`. Required if there are `Nan` in observed.
    """

    def __init__(self,
                 generator: Callable,
                 observed: SeriesComponent,
                 name: str = 'likelihood',
                 group: Union[str, Tuple[str, ...]] = 'obs',
                 nan_param_mask: Dict[str, Any] = None,
                 nan_data_mask: Any = None,
                 **kwargs: Any):
        if np.any(np.isnan(observed.values)):
            if nan_param_mask is None or nan_data_mask:
                raise ValueError(
                    'Both nan_param_mask and nan_data_mask must be defined when there is Nan in the observed data.')

            warnings.warn('Observables contains NaN values, these will be ignored in likelihood computation.',
                          UserWarning)

            components = {}

            for k, v in kwargs.items():
                if k not in nan_param_mask:
                    raise ValueError('Mask values for all parameters must be defined in nan_param_mask')

                var_mask = nan_param_mask[k] if isinstance(nan_param_mask, dict) else nan_param_mask

                component = GroupComponent(group=group, name=k + '_masked')
                for m in np.argwhere(~np.isnan(observed.values)).flatten():
                    component.add(m, v)
                for m in np.argwhere(np.isnan(observed.values)).flatten():
                    component.add(m, FixedValueComponent(value=var_mask, name=f'{k}_{m}_mask'))
                components[k] = component

            non_nandata = np.array([o if not np.isnan(o) else nan_data_mask for o in observed.values])
            components['observed'] = SeriesComponent(non_nandata, name='masked_' + observed.name)
        else:
            components = kwargs
            components['observed'] = observed

        super().__init__(generator, name, group, **components)



from sakkara.relation.atomicnode import AtomicNode
from typing import Set, Any
import numpy as np


class CellNode(AtomicNode):
    def __init__(self, name: Any, parents: Set[AtomicNode]):
        super().__init__(name, np.array(list()), parents)



from sakkara.model.composable.group.component import GroupComponent
from sakkara.model.composable.hierarchical.likelihood import Likelihood
from sakkara.model.composable.hierarchical.distribution import DistributionComponent
from sakkara.model.composable.hierarchical.deterministic import DeterministicComponent
from sakkara.model.fixed.base import FixedValueComponent
from sakkara.model.fixed.data import SeriesComponent, data_components
from sakkara.model.function import FunctionComponent
from sakkara.model.utils import build



from abc import ABC
from typing import Any, Dict, Union, Tuple, Optional, Collection

import pytensor.tensor as pt

from sakkara.model.base import ModelComponent
from sakkara.model.composable.base import Composable, T
from sakkara.relation.groupset import GroupSet


class HierarchicalComponent(Composable[str, T], ABC):
    """
    Base class for hierarchical components, i.e., components that may be defined on one or several column, composed
    by other :class:`ModelComponent` objects


    :param name: Name of the corresponding variable to register in PyMC.
    :param group: Group of which the component is defined for.
    :param members: Subset of members of column the component is defined for.
    :param subcomponents: Dict of underlying :class:`ModelComponent` objects.
    """
    def __init__(self, name: Optional[str], group: Optional[Union[str, Tuple[str, ...]]],
                 members: Optional[Collection[Any]], subcomponents: Dict[str, T]):
        super().__init__(name, group, members, subcomponents)

    def __getitem__(self, item: Any) -> ModelComponent:
        return self.subcomponents[item]

    def prebuild(self, groupset: GroupSet) -> None:
        self.build_components(groupset)

    def get_built_components(self) -> Dict[str, pt.Variable]:
        built_components = {}
        for key, comp in self.subcomponents.items():
            built_components[key] = comp.variable[self.node.map_to(comp.node)]

            if self.member_nodes.size < self.group_node.get_members().ravel().size:
                built_components[key] = built_components[key][self.node.member_subset(self.member_nodes)]

        return built_components





import abc
import time
from functools import cache
from typing import Set, Any, Tuple, Callable, Union

import numpy as np
import numpy.typing as npt


class Node:
    """
    Abstract class for relational nodes
    """

    @abc.abstractmethod
    def get_key(self) -> Tuple[Any, ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def is_parent_to(self, other: 'Node') -> bool:
        """
        Test if another node is a child of the node

        Parameters
        ----------
        other : Node
            Other node to test

        Returns
        -------
        True if the node is parent to other, False otherwise

        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_members(self) -> npt.NDArray['Node']:
        """
        Get all nodes linked as members of the node
        """

    @abc.abstractmethod
    def get_parents(self) -> Set['Node']:
        """
            Get all nodes linked as parents of the node
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_children(self) -> Set['Node']:
        raise NotImplementedError

    @abc.abstractmethod
    def get_twins(self) -> Set['Node']:
        raise NotImplementedError

    @abc.abstractmethod
    def is_twin_to(self, other: 'Node') -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def representation(self):
        """
            Returns the set of nodes that represents the information level of the group
        """
        raise NotImplementedError

    def get_mapping(self, other: 'Node', match_fct: Callable) -> Tuple[npt.NDArray[int], ...]:
        raveled_mapping = np.empty(self.get_members().ravel().shape, dtype=int)
        raveled_self_members = self.get_members().ravel()
        raveled_other_members = other.get_members().ravel()

        for raveled_self_member_index, self_member in enumerate(raveled_self_members):
            raveled_index = np.argmax(match_fct(self_member, raveled_other_members))
            raveled_mapping[raveled_self_member_index] = raveled_index

        return tuple(map(lambda x: x.reshape(self.get_members().shape).tolist(),
                         np.unravel_index(raveled_mapping, other.get_members().shape)))

    def map_to(self, other: 'Node') -> Union[Tuple[npt.NDArray[int], ...], slice]:
        """
        Get a mapping from other group to this group.

        Parameters
        ----------
        other: Group to get mapping from

        Returns
        -------
        List of the member nodes of other, in an order to be mapped to the correct member of this group (or a pair group)
        """
        if self == other:
            return slice(None)
        if self.representation() == other.representation() or other.is_twin_to(self) or self.is_twin_to(other):
            return self.get_mapping(other,
                                    np.vectorize(lambda om, sm: om.representation() == sm.representation()))

            if self.is_parent_to(other):
                raise ValueError('Mapping from parent to child is not applicable.')

        if not other.is_parent_to(self):
            raise ValueError('Mapping between unrelated nodes is not applicable.')

        # Other node is parent to self
        return self.get_mapping(other, np.vectorize(
            lambda sm, om: len(om.representation().get_nodes().intersection(sm.get_parents()))))

    def member_subset(self, subset: npt.NDArray['Node']) -> Tuple[npt.NDArray[int], ...]:
        vec_repr = np.vectorize(lambda a, b: a.representation() == b.representation())
        representation_match = np.vectorize(lambda node: np.any(vec_repr(node, subset)))

        vec_p2c = np.vectorize(lambda parent, child: parent.is_parent_to(child))
        parent_match = np.vectorize(lambda node: np.any(vec_p2c(subset, node)))

        member_node_match = np.vectorize(
            lambda member: np.logical_or(representation_match(member), parent_match(member)))

        raveled_member_indices = np.argwhere(member_node_match(self.get_members().ravel())).squeeze()

        return np.unravel_index(raveled_member_indices, shape=self.get_members().shape)

    @abc.abstractmethod
    def reduced_repr(self) -> 'Node':
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError



from abc import ABC
from typing import Optional, Union, Tuple

import pymc as pm

from sakkara.model.base import ModelComponent
from sakkara.model.composable.base import T
from sakkara.model.composable.hierarchical.base import HierarchicalComponent


class DeterministicComponent(HierarchicalComponent[T], ABC):
    """
    Wrapper for :class:`pymc.Deterministic`

    :param name: Name that will be applied to the :class:`pymc.Deterministic` object.
    :param component: :class:`ModelComponent` whose corresponding PyMC variable wil be wrapped into :class:`pymc.Deterministic`
    :param group: Group of which the component is defined for.
    """

    def __init__(self, name: str, component: ModelComponent, group: Optional[Union[str, Tuple[str, ...]]] = None):
        super().__init__(name, group, None, subcomponents={'var': component})

    def build_variable(self) -> None:
        self.variable = pm.Deterministic(self.name, **self.get_built_components(), dims=self.dims())



from abc import ABC
from typing import Set, Any, Tuple

from sakkara.relation.node import Node
from sakkara.relation.representation import Representation
import numpy.typing as npt


class AtomicNode(Node, ABC):
    """
        Class for a single instance node.
    """

    def __init__(self, name: Any, members: npt.NDArray['AtomicNode'], parents: Set['AtomicNode']):
        """

        Parameters
        ----------
        name: Name of group
        parents: All (atomic) immediate and non-immediate parents to this atomic group.
        """
        self.name = name
        self.members = members
        self.parents = set(parents)
        self.children = set()
        self.twins = set()

    def get_key(self) -> Tuple[Any, ...]:
        return self.name,

    def __str__(self):
        return str(self.name)

    def get_members(self) -> npt.NDArray[Node]:
        return self.members

    def add_child(self, child: 'AtomicNode'):
        self.children.add(child)

    def get_children(self) -> Set[Node]:
        return self.children

    def is_parent_to(self, other: Node) -> bool:
        return self in other.get_parents()

    def get_parents(self) -> Set[Node]:
        return self.parents

    def representation(self) -> Representation:
        return Representation({self})

    def reduced_repr(self) -> 'Node':
        return self

    def add_twin(self, twin: 'AtomicNode'):
        self.twins.add(twin)

    def get_twins(self) -> Set[Node]:
        return self.twins

    def is_twin_to(self, other: Node) -> bool:
        return self in other.get_twins()

    def __lt__(self, other) -> bool:
        return str(self) < str(other)

    def __len__(self) -> int:
        return 1




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sakkara',
 'sakkara.model',
 'sakkara.model.composable',
 'sakkara.model.composable.group',
 'sakkara.model.composable.hierarchical',
 'sakkara.model.fixed',
 'sakkara.relation']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23,<2.0',
 'pandas>=1.4,<2.0',
 'pymc>=5.0.0,<6.0.0',
 'pytensor>=2.9.1,<3.0.0']

setup_kwargs = {
    'name': 'sakkara',
    'version': '0.3.0',
    'description': 'Tools for fast creation of PyMC models',
    'long_description': 'None',
    'author': 'Henrik Håkansson',
    'author_email': 'henrik.hakansson@fcc.chalmers.se',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import operator
from abc import ABC
from typing import Any

from sakkara.model.base import ModelComponent
from sakkara.model.function import FunctionComponent


class MathOpBase(ModelComponent, ABC):
    """
    Base class for common mathematical operations to be performed on ModelComponent objects
    """

    def __add__(self, other: Any) -> ModelComponent:
        return FunctionComponent.math_op(operator.add, self, other)

    def __radd__(self, other: Any):
        return FunctionComponent.math_op(operator.add, other, self)

    def __sub__(self, other: Any) -> ModelComponent:
        return FunctionComponent.math_op(operator.sub, self, other)

    def __rsub__(self, other: Any):
        return FunctionComponent.math_op(operator.sub, other, self)

    def __mul__(self, other: Any) -> ModelComponent:
        return FunctionComponent.math_op(operator.mul, self, other)

    def __rmul__(self, other: Any) -> ModelComponent:
        return FunctionComponent.math_op(operator.add, other, self)

    def __truediv__(self, other: Any) -> ModelComponent:
        return FunctionComponent.math_op(operator.truediv, self, other)

    def __rtruediv__(self, other: Any):
        return FunctionComponent.math_op(operator.truediv, other, self)

    def __neg__(self):
        return FunctionComponent(operator.neg, self)



from sakkara.relation.node import Node
from typing import Iterable
import numpy as np


class Representation:

    def __init__(self, nodes: set['Node', ...]):
        self.nodes = nodes
        self.updated = True
        self.before_reduce = len(self.nodes)

    def get_order(self):
        if self.updated:
            self.reduce()
        list_sed_nodes = list(self.nodes)
        list_sed_nodes.sort()
        return list_sed_nodes

    def reduce(self):
        self.nodes = {n_i for n_i in self.nodes if not np.any([n_i.is_parent_to(n_j) for n_j in self.nodes])}
        lnodes = list(self.nodes)
        has_twin = [[j for j in range(i+1, len(lnodes)) if lnodes[i].is_twin_to(lnodes[j])] for i in range(len(lnodes)-1)] + [[]]
        index = np.ones((len(lnodes)), dtype=bool)
        for i in range(len(lnodes)):
            if len(has_twin[i]) > 0 and index[i]:
                index[has_twin[i]] = False
        self.nodes = {n_i for i, n_i in enumerate(lnodes) if index[i]}
        self.updated = False

    def set_nodes(self, nodes: set[Node, ...]):
        self.nodes = nodes
        self.comp_expected_len()

    def get_nodes(self):
        if self.updated:
            self.reduce()
        return self.nodes

    def __eq__(self, other: 'Representation') -> bool:
        if self.updated:
            self.reduce()
        nodes_eq = self.nodes == other.nodes
        if not nodes_eq:
            remaining_nodes = list(self.nodes.union(other.nodes) - self.nodes.intersection(other.nodes))
            n_nodes = len(remaining_nodes)
            matches = np.zeros((n_nodes, n_nodes), dtype=bool)
            for i in range(n_nodes-1):
                node_matches = [remaining_nodes[i].is_twin_to(remaining_nodes[j]) for j in range(i, n_nodes)]
                matches[i, i:] = node_matches
                matches[i:, i] = node_matches

            if np.all(np.any(matches, axis=0)):
                nodes_eq = True
        return nodes_eq

    def __copy__(self):
        if self.updated:
            self.reduce()
        return Representation(self.nodes.copy())

    def __iter__(self):
        if self.updated:
            self.reduce()
        return iter(sorted(self.nodes))

    def __str__(self) -> str:
        if self.updated:
            self.reduce()
        return '{' + ', '.join(map(str, self.nodes)) + '}'

    def __len__(self) -> int:
        if self.updated:
            self.reduce()
        return len(self.nodes)

    def get_expected_len(self):
        return self.before_reduce

    def comp_expected_len(self):
        self.before_reduce = len(self.nodes)

    # Set Operations
    def add(self, element: Node) -> None:
        self.nodes.add(element)
        self.updated = True
        self.comp_expected_len()

    def update(self, *s: Iterable[Node]) -> None:
        self.nodes.update(*s)
        self.comp_expected_len()
        self.updated = True

    def discard(self, element: Node) -> None:
        self.nodes.discard(element)
        self.comp_expected_len()
        self.updated = True

    def pop(self) -> Node:
        self.updated = True
        node = self.nodes.pop()
        self.comp_expected_len()
        return node

    def remove(self, element: Node) -> None:
        self.nodes.remove(element)
        self.updated = True
        self.comp_expected_len()

    def clear(self) -> None:
        self.nodes.clear()
        self.updated = True
        self.before_reduce = 0

    def union(self, *s: Iterable[Node]) -> 'Representation':
        copy_rep = self.__copy__()
        copy_rep.set_nodes(self.nodes.union(*s))
        return copy_rep

    def intersection(self, *s: Iterable[Node]) -> 'Representation':
        copy_rep = self.__copy__()
        copy_rep.set_nodes(self.nodes.intersection(*s))
        return copy_rep

    def difference(self, *s: Iterable[Node]) -> 'Representation':
        copy_rep = self.__copy__()
        copy_rep.set_nodes(self.nodes.difference(*s))
        return copy_rep



from sakkara.relation.node import Node
from sakkara.relation.representation import Representation
from abc import ABC
from typing import Set, Any, Tuple
from functools import cache
import numpy.typing as npt
import numpy as np


class NodePair(Node, ABC):
    """
        Class for pair nodes, to allow for multi-node representations
    """

    def __init__(self, a: Node, b: Node, rep_nodes: 'Representation' = None):
        """

        Parameters
        ----------
        a: First group of pair
        b: Second group of pair
        """
        self.a = a
        self.b = b
        self.rep_nodes = rep_nodes

    def set_rep_nodes(self, rep_nodes):
        self.rep_nodes = rep_nodes

    def get_key(self) -> Tuple[Any, ...]:
        return self.a.get_key() + self.b.get_key()

    def parent_child_member_pairs(self) -> npt.NDArray['NodePair']:
        """
        Member initialization if a and b has a parent-child relation
        Parameters

        Returns
        -------
        List of members, size same as member of child group
        """
        parent, child = (self.a, self.b) if self.a.is_parent_to(self.b) else (self.b, self.a)
        members = np.empty(child.get_members().shape, dtype=object)
        raveled_members = members.ravel()
        raveled_parents = parent.get_members().ravel()
        for i, child_member in enumerate(child.get_members().ravel()):
            # Find the member of parent group that is parent to this child member
            parent_member = next(filter(lambda p: p.is_parent_to(child_member), raveled_parents))
            raveled_members[i] = NodePair(parent_member, child_member)
        return np.array(members)

    def same_level_member_pairs(self) -> npt.NDArray['NodePair']:
        """
        Member initialization when a and b are not related. Creates a pair member for each pair of (member a, member b)
        where both member_a and member_b must be linked to the same parent member of a common parent of groups a and b.
        If group a and b have no common parent, the pair members will be the carthesian product of members of a and the
        members of b.
        Returns
        -------
        List of members, size dependent on relations to common parent
        """

        unraveled_a = self.a.get_members().ravel()
        unraveled_b = self.b.get_members().ravel()

        semi_raveled = np.empty((len(unraveled_a), len(unraveled_b)), dtype=object)
        for i, a_member in enumerate(unraveled_a):
            for j, b_member in enumerate(unraveled_b):
                semi_raveled[i, j] = NodePair(a_member, b_member)

        return semi_raveled.reshape(self.a.get_members().shape + self.b.get_members().shape)

    def get_members(self) -> npt.NDArray[Node]:
        """
        Init the member list, varying strategies depending on relation of a and b.
        Returns
        -------
        List of OrderedGroupPair, where a element of each OrderedGroupPair object corresponds to a member of the a group
        and b corresponds to a member of the b group.
        """
        if len(self.a.get_members()) == 0 and len(self.b.get_members()) == 0:
            return np.array(list())

        # Check if a and b has the same representation
        if self.a.representation() == self.b.representation():
            # a and b have the same representation nodes, possibly different order
            return self.a.get_members()

        # Check parent-child relationship
        if self.a.is_parent_to(self.b) or self.b.is_parent_to(self.a):
            # a and b have a parent-child relation
            return self.parent_child_member_pairs()

        # a and b are unrelated
        return self.same_level_member_pairs()

    def is_parent_to(self, other: Node) -> bool:
        if self.representation() == other.representation():
            return False
        if self.a.is_parent_to(self.b):
            return self.b.is_parent_to(other)
        if self.b.is_parent_to(self.a):
            return self.a.is_parent_to(other)
        a_par_other = self.a.is_parent_to(other)
        b_par_other = self.b.is_parent_to(other)
        a_twin_other = self.a.is_twin_to(other)
        b_twin_other = self.b.is_twin_to(other)
        return (a_par_other and b_twin_other) or (b_par_other and a_twin_other) or (a_par_other and b_par_other)

    def get_parents(self) -> Set[Node]:
        parents = set().union(*map(lambda g: g.get_parents(), self.representation()))

        if len(self.representation()) > 1:
            parents = parents.union(self.representation())

        return parents

    def get_children(self) -> Set[Node]:
        return set().intersection(*map(lambda g: g.get_children(), self.representation()))

    def get_twins(self) -> Set['Node']:
        twins = set().union(*map(lambda g: g.get_twins(), self.representation()))

        return twins

    def is_twin_to(self, other: Node) -> bool:
        if self.representation().get_nodes() == other.representation().get_nodes():
            return False
        return self.a.is_twin_to(other) and self.b.is_twin_to(other)

    # @cache
    def representation(self) -> Representation:
        if self.rep_nodes is None:
            if self.b.is_parent_to(self.a):
                self.rep_nodes = self.a.representation()
                return self.rep_nodes

            if self.a.is_parent_to(self.b):
                self.rep_nodes = self.b.representation()
                return self.rep_nodes

            self.rep_nodes = self.a.representation().union(self.b.representation())
        return self.rep_nodes

    def reduced_repr(self) -> Node:

        if self.representation() == self.a.representation() and self.a.__len__() == self.representation().__len__():
            return self.a.reduced_repr()
        if self.representation() == self.b.representation() and self.b.__len__() == self.representation().__len__():
            return self.b.reduced_repr()

        if self.__len__() != self.representation().__len__():
            return generate_nodepair(self.representation())
        return self

    # @cache
    def __str__(self) -> str:
        return '_'.join(map(str, self.representation()))

    def __len__(self) -> int:
        return len(self.a) + len(self.b)


def generate_nodepair(nodes: Representation):
    if len(nodes) == 1:
        return list(nodes)[0]
    else:
        lnodes = list(nodes.get_nodes())
        nodepair = NodePair(lnodes[0], lnodes[1])
        for node in lnodes[2:]:
            nodepair = NodePair(nodepair, node)
        nodepair.set_rep_nodes(nodes)
    return nodepair
